#include <algorithm>
#include <fstream>
#include <locale>
#include <memory>
#include <sstream>
#include <vector>

#include "crutil.hh"
#include "crinfo.hh"
#include <crdb.h>

#define ADS_BASE "https://ui.adsabs.harvard.edu/abs/"

namespace _CRDATASET {

std::shared_ptr<CRData> parse_data_crd(std::istream&, _CRUTIL::StreamParser&);
std::shared_ptr<CRData> parse_data_usine(std::istream&, _CRUTIL::StreamParser&);

std::string parse_link(const std::string& hint) {
    if (_CRUTIL::is_substr_i(hint, "http")) {
        return hint;
    } else if (_CRUTIL::is_substr_i(hint, "doi:")) {
        return "https://doi.org/" + hint.substr(4);
    } else if (_CRUTIL::is_substr_i(hint, "ads:")) {
        return "https://ui.adsabs.harvard.edu/abs/" + hint.substr(4);
    } else if (_CRUTIL::is_substr_i(hint, "arxiv:")) {
        return "https://arxiv.org/abs/" + hint.substr(6);
    } else {
        return "https://inspirehep.net/literature?q=" + hint;
    }
}

void after_burn(CRData& data) {
    data.use_rigidity = _CRUTIL::is_substr_i(data.x_quantity, "r");
    get_info(data);
    // TODO: time windows
}

}; // namespace _CRDATASET

CRDataset::CRDataset(const std::string& filename) { parse_file(filename); }

void CRDataset::parse_stream(std::istream& is, const std::string& filename) {

    enum { automatic, crd, usine /* , galprop */ } fileformat = automatic;

    _CRUTIL::StreamParser fp(filename);

    const auto& line = fp.current_line();
    fp.next_line(is, fp.skip::empty);
    while (is) {

        // determine fileformat
        if (fileformat == automatic) {
            if (line.substr(0, 4) == "@exp") {
                fileformat = crd;
            } else if (_CRUTIL::is_substr_i(line, "# Format: USINE code")) {
                // } else if (line.substr(0, 20) == "# Format: USINE code") {
                fileformat = usine;
            } else { // undetermined, read next line
                fp.next_line(is, fp.skip::empty);
                continue;
            }
        }

        switch (fileformat) {
        case crd: {
            this->push_back(_CRDATASET::parse_data_crd(is, fp));
            break;
        }
        case usine: {
            this->push_back(_CRDATASET::parse_data_usine(is, fp));
            break;
        }
        default: { // no way
            throw std::logic_error("unknown format of file '" + filename + "'");
        }
        }
    }
}

void CRDataset::parse_file(const std::string& filename) {
    std::ifstream ifs(filename);
    parse_stream(ifs, filename);
}

/** TEMPORARY MACRO **/
#define _GET_STRING_OPERAND(dest, join) \
    ss >> dest;                         \
    while (ss >> operand)               \
        dest += join + operand;

std::shared_ptr<CRData> _CRDATASET::parse_data_crd(std::istream& is,
                                                   _CRUTIL::StreamParser& fp) {
    auto pdata = std::make_shared<CRData>();
    auto& data = *pdata;
    std::vector<std::vector<double>> table;

    const auto& line = fp.current_line();
    std::string directive, operand;
    {
        std::stringstream ss(line);
        ss >> directive;
        if (directive == "@exp") {
            _GET_STRING_OPERAND(data.experiment, " ");
        } else {
            fp.panic<std::logic_error>("this should never happen");
        }
    }

    while (fp.next_line(is, fp.skip::both), is) {

        std::stringstream ss(line);

        double x, y, x_lower, x_upper, e;

        if (ss >> x) { // Try to get numbers
            if (!(ss >> x_lower >> x_upper >> y >> e)) {
                fp.panic(
                    "at least 5 real numbers are required for a data line");
            } else {
                if (x <= 0) { // use bin boundary to determine x
                    // TODO: support linear x
                    x = std::sqrt(x_lower * x_upper);
                } else { // for those provide bin info
                    if (x_lower <= 0) {
                        x_lower      = x;
                        data.has_bin = false;
                    }
                    if (x_upper <= 0) {
                        x_upper      = x;
                        data.has_bin = false;
                    }
                }
                table.push_back({x, x_lower, x_upper, y, e});
                while (ss >> e) {
                    table.back().push_back(e);
                }
            }
        } else if (ss.clear(), ss >> directive) { // Try to get string
            // if (directive.empty() || directive[0] == '#') { // commented
            //     continue;
            // } else
            if (directive[0] != '@') { // ill formatted
                fp.panic();

            } else if (directive == "@exp" || directive == "@th") {
                // terminate the current section
                break;

            } else if (directive == "@time") {
                _GET_STRING_OPERAND(data.time, ";");
                // if (!(ss >> data.time_begin >> data.time_end)) {
                //     ps.panic(
                //         "require 2 date operand YYYYMMDD after '@time'");
                // }

                // TODO: parse windows

            } else if (directive == "@cite") {
                _GET_STRING_OPERAND(data.cite, "");

            } else if (directive == "@link") {
                _GET_STRING_OPERAND(data.link, "");
                data.link = _CRDATASET::parse_link(data.link);

            } else if (directive == "@x") {
                _GET_STRING_OPERAND(data.x_quantity, " ");

            } else if (directive == "@x_unit") {
                _GET_STRING_OPERAND(data.x_unit, " ");

            } else if (directive == "@y") {
                _GET_STRING_OPERAND(data.y_quantity, " ");

            } else if (directive == "@y_unit") {
                _GET_STRING_OPERAND(data.y_unit, " ");
                if (data.y_unit == "none" || data.y_unit == "1") {
                    data.y_unit = "";
                }

            } else if (directive == "@errtype") {
                ss >> operand;
                data.skewed_error = operand == "skewed";

            } else if (directive == "@scale") {
                (ss >> data.scale) || (data.scale = 0.);

            } else if (directive == "@phi") {
                (ss >> data.phi) || (data.phi = 0);

            } else if (directive == "@solar_dist") {
                (ss >> data.solar_dist) || (data.solar_dist = 1);

            } else if (directive == "@flux_num") {
                // TODO: redundant since we have @y
            } else if (directive == "@flux_den") {
                // TODO: redundant since we have @y
            } else {
                // warn unused directive
            }
        } else { // empty action for empty line
        }
    } // while (fp.next_line(is, fp.skip::both), is)

    if (!table.size()) {
        // warn empty data
        fp.panic("no data point read");
        // return data;
    }

    // sort data points by x_lower
    std::sort(table.begin(), table.end(),
              [](const std::vector<double>& d1, const std::vector<double>& d2) {
                  return d1[1] < d2[1];
              });

    // store table into data object
    data._data.resize(table.size() * sizeof(CRData::DataPoint) /
                      sizeof(double));
    for (size_t i = 0; i < table.size(); ++i) {
        auto& dp       = data[i];
        const auto& tb = table[i];
        dp.x           = tb[0];
        dp.x_lower     = tb[1];
        dp.x_upper     = tb[2];
        dp.y           = tb[3];
        if (data.skewed_error && tb.size() > 5) {
            dp.ey1 = tb[4];
            dp.ey2 = tb[5];
        } else {
            if (tb.size() == 5) {
                dp.ey2 = dp.ey1 = tb[4];
            } else {
                double e = 0;
                for (size_t j = 4; j < tb.size(); ++j) {
                    e += tb[j] * tb[j];
                }
                dp.ey2 = dp.ey1 = std::sqrt(e);
            }
        }
    }

    // data.label =
    //     _CRDATASET::format("%s(%s-%s)", data.experiment.c_str(),
    //                        data.time_begin.c_str(), data.time_end.c_str());
    data.label = data.experiment;


    _CRDATASET::after_burn(data);
    return pdata;
}

#undef _GET_STRING_OPERAND

std::shared_ptr<CRData>
_CRDATASET::parse_data_usine(std::istream& is, _CRUTIL::StreamParser& fp) {
    auto pdata       = std::make_shared<CRData>();
    auto& data       = *pdata;
    const auto& line = fp.current_line();

    std::vector<std::vector<double>> table;

    // not read since the first line may already been read in previous cycle
    while (is) {
        if (line.empty() || line[0] == '#') {
            fp.next_line(is, fp.skip::both);
            continue;
        }

        std::string yquant, subexp, xquant;
        double x, x_lower, x_upper, y, estat1, estat2, esyst1, esyst2;
        std::string ads_id;
        double phi, dist;
        std::string datimes;
        bool is_upper_limit;

        std::stringstream ss(line);
        if (!(ss >> yquant          // Col.1
              >> subexp             // Col.2
              >> xquant             // Col.3
              >> x                  // Col.4
              >> x_lower >> x_upper // Col.5-6
              >> y                  // Col.7
              >> estat1 >> estat2   // Col.8-9
              >> esyst1 >> esyst2   // Col.10-11
              >> ads_id             // Col.12
              >> phi >> dist        // Col.13-14
              >> datimes            // Col.15
              >> is_upper_limit     // Col.16
              )) {
            fp.panic();
        }

        // remember not to store anything before confirming we are dealing the
        // same data
        if (data.y_quantity.empty() || data.experiment.empty()) {
            data.y_quantity = yquant;
            data.experiment = subexp;

            // parse experiment time
            // input format = yyyy/mm/dd-hhmmss:yyyy/mm/dd-hhmmss;...
            //                0....5..8........18...23.26........36
            // store format = yyyymmdd-yyyymmdd;...
            //                0........9.......18
            std::string windows_str;
            for (size_t i = 0; i + 34 < datimes.size(); i += 36) {
                windows_str += datimes.substr(i, 11);
                windows_str += datimes.substr(i + 18, 10);
                windows_str += datimes.substr(i + 35, 1);
            }
            // std::erase(windows_str, '/'); // -std=c++2a
            (void)std::remove(windows_str.begin(), windows_str.end(), '/');

            data.time = windows_str;
            // TODO: parse windows

            // other info
            data.x_quantity = xquant;
            data.cite       = ads_id;
            // data.link       = ADS_BASE + ads_id;
            data.link       = _CRDATASET::parse_link("ads:" + ads_id);
            data.phi        = phi;
            data.solar_dist = dist;
        } else { // check if data changed
            if (data.y_quantity != yquant || data.experiment != subexp) {
                // do not proceed reading, finalize what have been read
                break; // while (fp)
            }
        }

        data.has_bin = data.has_bin && x_upper > x_lower;

        // store values into table
        double e1, e2;
        if (is_upper_limit) {
            // XXX definition not clear, may be incorrect
            e1 = e2 = y;
            y       = 0;
        } else {
            e1 = std::sqrt(estat1 * estat1 + esyst1 * esyst1);
            e2 = std::sqrt(estat2 * estat2 + esyst2 * esyst2);
        }

        table.push_back({x, x_lower, x_upper, y, e1, e2});

        // repeat for the next line
        // ps.getline(is);
        fp.next_line(is, fp.skip::both);

    } // while (is)

    // store table into data object
    data._data.resize(table.size() * sizeof(CRData::DataPoint) /
                      sizeof(double));
    for (size_t i = 0; i < table.size(); ++i) {
        auto& dp       = data[i];
        const auto& tb = table[i];
        dp.x           = tb[0];
        dp.x_lower     = tb[1];
        dp.x_upper     = tb[2];
        dp.y           = tb[3];
        dp.ey1         = tb[4];
        dp.ey2         = tb[5];
    }

    // TODO: x_uint, y_unit

    _CRDATASET::after_burn(data);
    return pdata;
}

std::shared_ptr<CRData> CRDataset::find_exact(const std::string& quantity,
                                              const std::string& exp) const {
    std::shared_ptr<CRData> found = nullptr;
    for (const auto& e : *this) {
        if (e->experiment == exp && quantity == e->y_quantity) {
            if (found) {
                throw std::runtime_error("ambiguous '" + quantity +
                                         "' data found for '" + exp + "'");
            }
            found = e;
        }
    }
    return found;
}

std::shared_ptr<CRData> CRDataset::find_match(const std::string& quantity,
                                              const std::string& exp) const {
    std::shared_ptr<CRData> found = nullptr;
    size_t eff_len = -1;
    for (const auto& e : *this) {
        if (_CRUTIL::is_substr_i(e->experiment, exp) &&
            quantity == e->y_quantity) {
            size_t new_eff_len = e->experiment.find('(');
            if (new_eff_len == std::string::npos) {
                new_eff_len = e->experiment.size();
            }
            if (found && new_eff_len >= eff_len) {
                // use minimum matched entry
                continue;
            }
            found = e;
            eff_len = new_eff_len;
        }
    }
    return found;
}

CRData CRData::from_crd(const std::string& filename) {
    _CRUTIL::StreamParser fp(filename);
    std::ifstream ifs(filename);
    fp.next_line(ifs, fp.skip::both);
    return *_CRDATASET::parse_data_crd(ifs, fp);
}
CRData CRData::from_usine(const std::string& filename) {
    _CRUTIL::StreamParser fp(filename);
    std::ifstream ifs(filename);
    fp.next_line(ifs, fp.skip::both);
    return *_CRDATASET::parse_data_usine(ifs, fp);
}
