#include <curl/curl.h>
#include <filesystem>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

#ifndef WITHOUT_ROOT
#define WITHOUT_ROOT // avoid root error with c++17
#include <crdb.h>
#undef WITHOUT_ROOT
#else
#include <crdb.h>
#endif

static size_t WriteCallback(void* contents, size_t size, size_t nmemb,
                            void* userp) {
    const size_t realsize = size * nmemb;
    auto& osv             = *((std::vector<std::ostream*>*)userp);
    for (auto os : osv) {
        os->write((char*)contents, realsize);
    }
    return realsize;
}

CRDataset CRRequest::fetch(const std::string& cache_dir) const {
    CRDataset ds;
    fetch(ds, cache_dir);
    return ds;
}

CRDataset& CRRequest::fetch(CRDataset& ds, const std::string& cache_dir) const {
    std::stringstream ss;
    fetch(ss, cache_dir);
    ds.parse_stream(ss);
    return ds;
}

std::ostream& CRRequest::fetch(std::ostream& os,
                               const std::string cache_dir) const {

    // ensure that numerator is set
    if (num.empty()) {
        throw std::logic_error("unspecified numerator CR species");
    }

    std::string num_used, den_used;
    {
        auto islash = num.find('/');
        if (islash != num.npos) {
            if (!den.empty()) {
                throw std::runtime_error(
                    "duplicated specification of denominator: " + num + '/' +
                    den);
            }
            den_used = num.substr(islash + 1);
            num_used = num.substr(0, islash);
        } else {
            den_used = den;
            num_used = num;
        }
    }

    std::vector<std::ostream*> os_list = {&os};
    std::ofstream ofs; // must live long enough till writing

    if (!cache_dir.empty()) {
        // construct cache filename
        std::string filename =
            num_used + '_' + (den_used.empty() ? "" : den_used + '_') +
            energy_type + '_' + static_cast<char>(accuracy) + ".usine";

        std::filesystem::path cache_path(cache_dir);
        if (!std::filesystem::is_directory(cache_path)) {
            std::filesystem::create_directories(cache_path); // with exceptions
        }

        auto file_path = cache_path / filename;
        if (std::filesystem::exists(file_path)) {
            // already fetched
            std::ifstream ifs(file_path.string());
            return os << ifs.rdbuf();
        } else {
            ofs.open(file_path.string());
            os_list.push_back(&ofs);
        }
    }

    // url encode '+'
    {
        size_t iplus;
        while ((iplus = num_used.find('+')) != num_used.npos) {
            num_used.replace(iplus, 1, "%2B");
        }
        while ((iplus = den_used.find('+')) != den_used.npos) {
            den_used.replace(iplus, 1, "%2B");
        }
    }

    // construct url
    const std::string url =
        // std::string("http://lpsc.in2p3.fr/crdb/rest.php?format=usine&num=") +
        std::string(
            "https://lpsc.in2p3.fr/crdb/_dialog_result.php?format=usine&num=") +
        num_used + "&den=" + den_used + "&energy_type=" + energy_type +
        "&energy_convert_level=" + static_cast<char>(accuracy);

    // allocate curl instance
    CURL* curl = curl_easy_init();

    // specify url to fetch
    curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
    // follow http redirect
    curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, true);
    // provide user defined context pointer, the list of ostream to write into
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, &os_list);
    // provide user defined writing callback function handling the context pointer
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteCallback);

    // perform requeset
    CURLcode code = curl_easy_perform(curl);
    curl_easy_cleanup(curl);

    if (code) {
        throw std::runtime_error("curl fetch failed: error code " +
                                 std::to_string(code));
    }

    return os;
}
